/*

  FIOLIB.C - I/O library for CPC with FutureOS
  To use with the Small-C compiler

  Adapted for FutureOS with enhancements by Stefan W. Stumpferl
  (previously Adapted from CP/M to AmsDOS by Juergen Weber)

  FutureOS 2004 by SWS - 12.02.2005
  AmsDOS   1993 by JGW
  CP/M     1984 by JRVZ

  This file has 1276 lines (test it!)

To compile:
-----------
   A>CC -M FIOLIB
   A>ZMAC FIOLIB=FIOLIB

FIOLIB.OBJ should be linked as FIRST object file!

NOTE: FIOLIB supports only a subset of the FutureOS system calls.
      But it should be easy to append some more.
*/

#include FIOLIB.H


#asm
global qmain
 call qcpcstart
 call qmain
 call qbyebye
 ld hl,0FE9DH ;TUR_E
 jp 0FF12H    ;ROM_D - back to FutureOS desktop
#endasm

int regs[4];
int stdin,stdout;


ccgo() { /* dummy, in case -M was forgotten */ }


cpcstart()
{
#asm

global qcpcstart

 pop hl
 ld  (retsto),hl ;RET addr.

#endasm

stdin=STDIN;
stdout=STDOUT;

#asm

 LD SP,09900H ;Stack for RECURSIVE C PROGS
 LD HL,09000H
 LD (heapend),HL

 LD   HL,0D60EH ;S80X25
 CALL 0FF00H    ;ROM_A

 ld hl,(retsto)
 jp (hl)

retsto: db 09,09

#endasm
}


byebye()
{
 puts("\nPress KEY to return to the Desktop.");
 getchar();

#asm
 POP  HL
 LD   SP,0BFFEH ;FutureOS stack
 PUSH HL
#endasm
}


oscall(adr,regpack)
int adr;
int *regpack; /* af,hl,de,bc */
{
#asm
 pop bc ;ret
 pop de ;regs
 pop hl ;adr

 push hl
 push de
 push bc

 ld a,0C3h
 ld (030h),a  ;rst 30h
 ld (031h),hl ;user

 push de
 ex (sp),ix

 ld l,(ix+0)
 ld h,(ix+1)

 push hl
 pop  af

 ld l,(ix+2)
 ld h,(ix+3)
 ld e,(ix+4)
 ld d,(ix+5)
 ld c,(ix+6)
 ld b,(ix+7)

 ex (sp),ix

 rst 30h ; execute os-call

 ex (sp),ix

 ld (ix+2),l
 ld (ix+3),h

 push af
 pop  hl

 ld (ix+0),l
 ld (ix+1),h
 ld (ix+4),e
 ld (ix+5),d
 ld (ix+6),c
 ld (ix+7),b

 pop de
 JP  CCSXT ; move A to HL & sign extend

#endasm
}


/* CPC specific functions */

getkey(n) int *n;
{
 regs[0]=0;
 oscall(WAIT_CHAR_KM,regs);
 *n=regs[0] >> 8;
}


mode(n) int n;
{
 regs[0]=n << 8;
 oscall(SET_MODE_SCR,regs);
}


cls()
{
 oscall(CLR_WIN_TXT,regs);
}


s80x25()
{
#asm
 call 0D60EH ;set screen to 80 columns and 25 lines
#endasm
}


s64x32()
{
#asm
 call 0D5A8H ;64 col * 32 lines
#endasm
}


pen0(n) int n;
{
 regs[0]=n << 8;
 oscall(OS_SET_0PEN,regs);
}


pen1(n) int n;
{
 regs[0]=n << 8;
 oscall(OS_SET_1PEN,regs);
}


pen2(n) int n;
{
 regs[0]=n << 8;
 oscall(OS_SET_2PEN,regs);
}


ink(p,c) int p,c;
{
 regs[1]=p;
 regs[2]=c;
 oscall(OS_SET_INK,regs);
}


locate(x,y) int x,y;
{
 regs[1]=y;
 regs[2]=x;
 oscall(OS_LOCATE,regs);
}


circle(r) int r;
{
 regs[0]=r << 8;
 oscall(OS_CIRCLE,regs);
}


plot(x,y) int x,y;
{
 regs[1]=y;
 regs[2]=x;
 oscall(OS_PLOT_XY,regs);
}


move(x,y) int x,y;
{
 regs[1]=y;
 regs[2]=x;
 oscall(MOVE_ABS_GRA,regs);
}


draw(x,y) int x,y;
{
 regs[1]=y;
 regs[2]=x;
 oscall(LINE_ABS_GRA,regs);
}


set_time(l,h) int h,l;
{
 regs[1]=l;  /* hl */
 regs[2]=h;  /* de */
 oscall(TIME_SET_KL,regs);
}


get_time(l,h) int *h,*l;
{
 oscall(TIME_GET_KL,regs);
 *l=regs[1];
 *h=regs[2];
}


lscr(s) char *s;
{
 regs[1]=s;
 oscall(OS_LD_SCR,regs);
}


fopen(nam,mod) char *nam,*mod;
{
 regs[1]=mod;
 regs[2]=nam;
 oscall(OUT_OPEN_CAS,regs);

 if (regs[0] & 1)       /* Carry set -> open ok */
    return(regs[0]>>8); /* Number of File 32-64 */
 else
    return(EOF); /* no Carry -> Error! */
}


fclose(f) int f;
{
 regs[0]=f << 8;
 oscall(IN_CLOSE_CAS,regs);
}


sfile(n) int n;
{
 regs[0]=n << 8;
 oscall(IN_OPEN_CAS,regs);
}


sfico(c) int c;
{
 regs[2]=c;
 oscall(OUT_CLOSE_CAS,regs);
}


cat()
{
 oscall(IN_CLOSE_CAS,regs); /* nur ein Dummy bis CAT implementiert */
}


err(s) char *s;
{
 puts("Error: ");
 puts(s);
}


putchar(c) char c;
{
 putc(c,stdout);
 return c;
}


putc(c,unit) char c; int unit;  /* write a character (to a file) */
{
 putb(c,unit);
 if(c=='\n')putb(LF,unit);      /* add LF after CR */
 return c;
}


putb(c,unit) char c; int unit;  /* write a byte (to a file) */
{
   regs[0]=c << 8;
   if(unit==STDOUT)
       {
       oscall(OUTPUT_TXT,regs);
 return c;
 }
   else {
       oscall(OUT_CHAR_CAS,regs);
       if ((regs[0] & 1)==0)
          return(EOF);
   }
   return(c);
}


gets(s) char *s;
{
#asm
 push bc
 push de
 push hl

 ld hl,45056
 ld (hl),0
 ld de,45057
 ld bc,00081

 ldir

 pop hl
 pop de
 pop bc
#endasm

 basedit(45056);
 strncpy(s,45056,81);
}


edst(s) char *s;
{
 strncpy(45056,s,81);
 basedit(45056);
 strncpy(s,45056,81);
}


basedit(b) char *b;
{
#asm

 CALL 0C4DBH ;WART_TS - wait until keyboard is free
 POP  DE
 POP  HL
 PUSH HL
 PUSH DE
 JP   39225

#endasm
}


getchar()
{
 return getc(stdin);
}


getc(unit) int unit;  /* get character (from file) */
{
 int c;
 while((c=getb(unit))==LF); /* discard LF */
 if (c==26)                 /* CP/M EOF? */
    return(EOF);
 return c;
}


getb(unit) int unit;  /* get byte from file */
{
 if(unit==STDIN)
   oscall(WAIT_CHAR_KM,regs);
 else {
   oscall(IN_CHAR_CAS,regs);
   if ((regs[0] & 1)==0)
      return(EOF);
 }
 return (regs[0]>>8);
}


#asm

; heap grows upwards (jgw)
;
; /* return address of a block of memory */
; alloc(b)
; int b;  /* # bytes desired */

QALLOC: POP HL ;return addr

 POP  DE ;block size
 PUSH DE
 PUSH HL

 LD  HL,(HEAPTOP) ;current top of heap
 EX  DE,HL
 ADD HL,DE        ;hl=new top of heap
 LD  (HEAPTOP),HL
 EX  DE,HL        ;hl=old top of heap
 RET


HEAPTOP: defw _END
heapend: defw 09000H

global QHEAPTOP
global Qheapend

; /* reset the top of heap pointer to addr* */
;
; free(addr)
; int addr;

QFREE: POP DE

 POP  HL ;addr
 PUSH HL
 PUSH DE
 LD   (HEAPTOP),HL
 RET


; /* return number of bytes between top of heap
; and end of TPA.  Remember that this includes
; the stack! */
;
; avail()

QAVAIL: LD HL,(heapend) ;end of TPA

 PUSH HL
 LD   HL,(HEAPTOP) ;top of heap
 JP   CCSUB        ;find (heapend)-HEAPTOP

#endasm


puts(buf) char *buf; /* print a null-terminated string */
{
 char c;
 while(c=*buf++) putchar(c);
}


upper(c) int c;  /* converts to upper case */
{
 if(c>='a')return c-32;
 return c;
}


#asm

;I/P a byte from a port 
;Function format: getp(n) where n is a port number

getp: POP DE ;Strip off return address

 POP  BC ;Get port number into BC
 PUSH BC
 PUSH DE
 IN   A,(C) ;Get input value into A
 JP   ccsxt ;Sign extend into HL and return


;OUT byte to port n 
;Function format: putp('A',n); or putp(value,n);

putp: POP DE ;Save return address

 POP  BC    ;Port no into BC
 POP  HL    ;Value for O/P
 OUT  (C),L
 PUSH HL    ;Restore stack
 PUSH BC
 PUSH DE
 RET

#endasm


/*
rand(n) returns an integer value between 0 and n 
If n is 0 the value will lie within +/- 32k
Based on suggestions in Hamming's book on
Numerical Methods for Scientists and Engineers 
A seed is multiplied by a large number causing an overflow.
*/

rand(n)
int n;
{
#asm
 LD A,R
 LD (RANSEED),A ;Set up random number seed
 JR RASE

RANSEED: defw 09999H

RASE: LD A,R

 LD (RANSEED+1),A
#endasm

 int *iptr,n1;
 iptr=RANSEED;
 n1=*iptr;
 n1=n1*2051;           /* Multiply by 8N+3 */
 if(n>=0)n++;else n--;
 *iptr=n1/2;           /* Avoid being trapped with even numbers */
 if(n==1) return n1;
 if(n1<0)n1=-n1;
 return (n1%n);
}


exit()
{
#asm
 LD SP,0BFFEH ;init FutureOS stack
 LD HL,0FE9DH ;TUR_E
 JP 0FF12H    ;ROM_D - exit to FutureOS
#endasm
}


#asm

; Fetch a single byte from the address in HL
; and sign extend into HL

CCGCHAR: LD A,(HL)
QARGC:
CCSXT: LD L,A

 RLCA
 SBC A,A
 LD  H,A
 RET


;Fetch integer from (HL+2)

CCCDR: INC HL
       INC HL

;Fetch a full 16-bit integer from the address in HL

CCCAR:
CCGINT: LD A,(HL)

 INC HL
 LD  H,(HL)
 LD  L,A
 RET


;Store a 16-bit integer in HL at the address in TOS

CCPINT: POP BC

 POP  DE
 PUSH BC

 LD  A,L
 LD  (DE),A
 INC DE
 LD  A,H
 LD  (DE),A
 RET


;Inclusive "or" HL and TOS into HL

CCOR: POP BC

 POP  DE
 PUSH BC

 LD A,L
 OR E
 LD L,A
 LD A,H
 OR D
 LD H,A

 RET


;Exclusive "or" HL and TOS into HL

CCXOR: POP BC

 POP  DE
 PUSH BC

 LD  A,L
 XOR E
 LD  L,A
 LD  A,H
 XOR D
 LD  H,A
 RET


;"And" HL and TOS into HL

CCAND: POP BC

 POP  DE
 PUSH BC

 LD  A,L
 AND E
 LD  L,A
 LD  A,H
 AND D
 LD  H,A
 RET


;Test if HL = TOS and set HL = 1 if true else 0

CCEQ: POP BC

 POP  DE
 PUSH BC
 CALL CCCMP

 RET Z
 DEC HL
 RET


;Test if TOS ~= HL

CCNE: POP BC

 POP  DE
 PUSH BC
 CALL CCCMP

 RET NZ
 DEC HL
 RET


;Test if TOS > HL (signed)

CCGT: POP BC

 POP  DE
 PUSH BC
 EX   DE,HL
 CALL CCCMP

 RET C
 DEC HL
 RET


;Test if TOS <= HL (signed)

CCLE: POP BC

 POP  DE
 PUSH BC
 CALL CCCMP

 RET Z
 RET C
 DEC HL
 RET


;Test if TOS >= HL (signed)

CCGE: POP BC

 POP  DE
 PUSH BC
 CALL CCCMP

 RET NC
 DEC HL
 RET


;Test if TOS < HL (signed)

CCLT: POP BC

 POP  DE
 PUSH BC
 CALL CCCMP

 RET C
 DEC HL
 RET


; Common routine to perform a signed compare
; of DE and HL
; This routine performs DE - HL and sets the conditions:
; Carry reflects sign of difference (set means DE < HL)
; Zero/non-zero set according to equality.

CCCMP: LD A,E

 SUB L
 LD  E,A
 LD  A,D
 SBC A,H
 LD  HL,1     ;preset true condition
 JP  M,CCCMP1
 OR  E        ;"OR" resets carry
 RET

CCCMP1: OR E

 SCF ;set carry to signal minus
 RET


;Test if TOS >= HL (unsigned)

CCUGE: POP BC

 POP  DE
 PUSH BC
 CALL CCUCMP

 RET NC
 DEC HL
 RET


;Test if TOS < HL (unsigned)

CCULT: POP BC

 POP  DE
 PUSH BC
 CALL CCUCMP
 RET  C
 DEC  HL
 RET


;Test if TOS > HL (unsigned)

CCUGT: POP BC

 POP  DE
 PUSH BC
 EX   DE,HL
 CALL CCUCMP

 RET C
 DEC HL
 RET


;Test if TOS <= HL (unsigned)

CCULE: POP BC

 POP  DE
 PUSH BC
 CALL CCUCMP

 RET Z
 RET C
 DEC HL
 RET


;Common routine to perform unsigned compare
;carry set if DE < HL
;zero/nonzero set accordingly

CCUCMP: LD A,D

 CP H
 JP NZ,CUCMP1
 LD A,E
 CP L

CUCMP1: LD HL,1

 RET


;Shift DE arithmetically right by HL and return in HL

CCASR: EX DE,HL

 DEC E
 RET M   ;7/2/82 jrvz
 LD  A,H
 RLA 
 LD  A,H
 RRA 
 LD  H,A
 LD  A,L
 RRA 
 LD  L,A
 JP  CCASR+1

;Shift TOS arithmetically left by HL and return in HL

CCASL: POP BC

 POP  DE
 PUSH BC
 EX   DE,HL

CCASL4: DEC E

 RET M      ; jrvz 7/2/82
 ADD HL,HL
 JP  CCASL4

;Subtract HL from TOS and return in HL

CCSUB: POP BC

 POP  DE
 PUSH BC

 LD  A,E
 SUB L
 LD  L,A
 LD  A,D
 SBC A,H
 LD  H,A
 RET


;Form the two's complement of HL

CCNEG: CALL CCCOM

 INC HL
 RET


;Form the one's complement of HL

CCCOM: LD A,H

 CPL 
 LD  H,A
 LD  A,L
 CPL 
 LD  L,A
 RET


;Multiply TOS by HL and return in HL

CCMULT: POP BC

 POP  DE
 PUSH BC

 LD B,H
 LD C,L
 LD HL,0

CCMLT1: LD A,C

 RRCA 
 JP  NC,CMLT2
 ADD HL,DE

CMLT2: XOR A

 LD  A,B
 RRA 
 LD  B,A
 LD  A,C
 RRA 
 LD  C,A
 OR  B
 RET Z
 XOR A
 LD  A,E
 RLA 
 LD  E,A
 LD  A,D
 RLA 
 LD  D,A
 OR  E
 RET Z
 JP  CCMLT1


;Divide DE by HL and return quotient in HL, remainder in DE

CCDIV: LD B,H

 LD   C,L
 LD   A,D
 XOR  B
 PUSH AF
 LD   A,D
 OR   A
 CALL M,CCDENEG
 LD   A,B
 OR   A
 CALL M,CCBCNEG
 LD   A,16
 PUSH AF
 EX   DE,HL
 LD   DE,0

CCDIV1: ADD HL,HL

 CALL CCRDEL
 JP   Z,CCDIV2
 CALL CCPBCDE
 JP   M,CCDIV2

 LD  A,L
 OR  1
 LD  L,A
 LD  A,E
 SUB C
 LD  E,A
 LD  A,D
 SBC A,B
 LD  D,A

CCDIV2: POP AF

 DEC  A
 JP   Z,CCDIV3
 PUSH AF
 JP   CCDIV1

CCDIV3: POP AF

 RET  P
 CALL CCDENEG
 EX   DE,HL
 CALL CCDENEG
 EX   DE,HL
 RET

CCDENEG: LD A,D

 CPL 
 LD  D,A
 LD  A,E
 CPL
 LD  E,A
 INC DE
 RET

CCBCNEG: LD A,B

 CPL 
 LD  B,A
 LD  A,C
 CPL
 LD  C,A
 INC BC
 RET

CCRDEL: LD A,E

 RLA 
 LD  E,A
 LD  A,D
 RLA
 LD  D,A
 OR  E
 RET

CCPBCDE: LD A,E

 SUB C
 LD  A,D
 SBC A,B
 RET

#endasm


hex(x) int x;
{
hexb(x>>8); hexb(x); putchar(' ');
}


hexb(x) int x;
{
hexn(x>>4); hexn(x);
}


hexn(x) int x;
{
x=x&15;
if(x<10) putchar(x+'0');
else putchar(x-10+'a');
}


/*
 * STRING FUNCTIONS FOR SMALL C
 *  BASED ON CORRESPONDING UNIX FUNCTIONS
 */

#asm

qstrcmp: POP BC

 POP  HL ;HL is s2
 POP  DE ;DE is s1
 PUSH DE
 PUSH HL
 PUSH BC

ccstr1: LD A,(DE) ;fetch *s1

 CP (HL)
 JP NZ,ccstr6 ;quit if *s1 != *s2
 OR A         ;check *s1 for zero

 INC DE
 INC HL
 JP  NZ,ccstr1 ;loop if *s1 != 0
 LD  HL,0
 RET

ccstr6: SUB (HL)

 JP ccsxt ;else return *s1-*s2

#endasm


/*
 * strncmp - compares two strings for at most n characters
 * and returns an integer >0, =0 or <0 as s is >t, =t or <t
 */

strncmp(s,t,n)
char *s,*t;
int n;
{
 while(n!=0 & *s==*t) {
  if( *s == 0 ) return 0;
  ++s;
  ++t;
  --n;
 }
 if(n) return *s - *t;
 return 0;
}


#asm

qstrcpy: POP HL

 POP  DE ;DE is from
 POP  BC ;BC is to
 PUSH BC
 PUSH DE
 PUSH HL
 LD  H,B ;return to
 LD  L,C

ccstr2: LD A,(DE)

 LD  (BC),A
 INC DE
 INC BC
 OR  A  ;test char for zero
 JP  NZ,ccstr2
 RET

#endasm


strncpy(tar,src,n) char *src,*tar;int n;
{
char *s,*t;
s=src;t=tar;
while (n--)
   if ((*tar++=*src++)=='\0') {
      break; }
*tar='\0';
}


#asm

qstrlen: POP HL

 POP  BC
 PUSH BC
 PUSH HL
 LD   HL,0

ccstr3: LD A,(BC)

 OR  A
 RET Z
 INC BC
 INC HL
 JP  ccstr3


qstrcat: POP HL

 POP  BC ;BC is from
 POP  DE ;DE is to
 PUSH DE
 PUSH BC
 PUSH HL
 LD  H,D ;return to
 LD  L,E

ccstr4: LD A,(DE)

 OR  A
 INC DE
 JP  NZ,ccstr4
 DEC DE

ccstr5: LD A,(BC)

 LD  (DE),A
 INC DE
 INC BC
 OR  A
 RET Z
 JP  ccstr5

#endasm


/*
 *  strncat - concatenate n bytes max from from to end of to
 *            to must be long enough
 */

strncat(to,from,n)
char *to,*from;
int n;
{
 char *temp;

 temp=to;
 --to;
 while(*++to);
 while(n--) {
  if(*to++=*from++) continue;
  return temp;
 }
 *to = 0;
 return temp;
}


/*
 * strchr - return pointer to first occurrence of c in string, else 0
 */

strchr(string,c)
char *string,c;
{
 while(*string) {
  if(*string == c) return string;
  ++string;
 }
 return 0;
}


/*
 *  strrchr - search string for rightmost occurrence of c
 *  returns pointer to rightmost c or NULL
 */

strrchr(string,c)
char *string,c;
{
 char *ptr;
 ptr = 0;
 while(*string) {
  if(*string == c) ptr = string;
  ++string;
 }
 return ptr;
}


tolower(s) char *s;
{
int c;
 while (c=*s) {
   if ((c>='A') & (c<='Z'))
       *s=c+'a'-'A';
   s++;
 }
}


toupper(s) char *s;
{
int c;
 while (c=*s) {
   if ((c>='a') & (c<='z'))
       *s=c+'A'-'a';
   s++;
 }
}
h enhancem